/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright held by original author
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

Author
    Bahram Haddadi, TUWards OpenFOAM.  All rights reserved

\*---------------------------------------------------------------------------*/

#include "regionCouplingFvPatchField.H"

// * * * * * * * * * * * * * * * * Private members  * * * * * * * * * * * * *//

template<class Type>
Foam::scalarField Foam::regionCouplingFvPatchField<Type>::
weights() const
{
    // Implement weights-based stabilised harmonic interpolation using magnitude
    // Algorithm:
    // 1) calculate magnitude of internal field and neighbour field
    // 2) calculate harmonic mean magnitude
    // 3) express harmonic mean magnitude as: mean = w*mOwn + (1 - w)*mNei
    // 4) Based on above, calculate w = (mean - mNei)/(mOwn - mNei)
    // 5) Use weights to interpolate values

    Field<Type> fOwn = this->patchInternalField();
    Field<Type> fNei = this->patchNeighbourField();

    scalarField mOwn = mag(fOwn);
    scalarField mNei = mag(fNei);
    scalarField mean = 2*(mOwn*mNei)/(mOwn + mNei + SMALL);

    scalarField weights(fOwn.size(), 0.5);

    scalar den;

    forAll (weights, faceI)
    {
        den = (mOwn[faceI] - mNei[faceI]);

        if (mag(den) > SMALL)
        {
            weights[faceI] = (mean[faceI] - mNei[faceI])/den;
        }
    }

    return weights;
}


// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

template<class Type>
Foam::regionCouplingFvPatchField<Type>::
regionCouplingFvPatchField
(
    const fvPatch& p,
    const DimensionedField<Type, volMesh>& iF
)
:
    coupledFvPatchField<Type>(p, iF),
    regionCoupledPatch_(refCast<const regionCoupledBaseFvPatch>(p)),
    nbrFieldName_(iF.name()),
    matrixUpdateBuffer_()
{}


template<class Type>
Foam::regionCouplingFvPatchField<Type>::
regionCouplingFvPatchField
(
    const regionCouplingFvPatchField<Type>& ptf,
    const fvPatch& p,
    const DimensionedField<Type, volMesh>& iF,
    const fvPatchFieldMapper& mapper
)
:
    coupledFvPatchField<Type>(ptf, p, iF, mapper),
    regionCoupledPatch_(refCast<const regionCoupledBaseFvPatch>(p)),
    nbrFieldName_(ptf.nbrFieldName_),
    matrixUpdateBuffer_()
{}


template<class Type>
Foam::regionCouplingFvPatchField<Type>::
regionCouplingFvPatchField
(
    const fvPatch& p,
    const DimensionedField<Type, volMesh>& iF,
    const dictionary& dict
)
:
    coupledFvPatchField<Type>(p, iF, dict),
    regionCoupledPatch_(refCast<const regionCoupledBaseFvPatch>(p)),
    nbrFieldName_(dict.lookup("nbrField")),
    matrixUpdateBuffer_()
{

    if (!isA<regionCoupledBase>(this->patch().patch()))
    {
        FatalErrorIn
        (
            "regionCouplingFvPatchField<Type>::"
            "regionCouplingFvPatchField<Type>\n"
            "(\n"
            "    const fvPatch& p,\n"
            "    const DimensionedField<Type, volMesh>& iF,\n"
            "    const dictionary& dict\n"
            ")\n"
        )   << "\n    patch type '" << p.type()
            << "' not type '" << regionCoupledBase::typeName << "'"
            << "\n    for patch " << p.name()
            << " of field " << this->dimensionedInternalField().name()
            << " in file " << this->dimensionedInternalField().objectPath()
            << exit(FatalError);
    }
}


template<class Type>
Foam::regionCouplingFvPatchField<Type>::
regionCouplingFvPatchField
(
    const regionCouplingFvPatchField<Type>& ptf
)
:
    coupledFvPatchField<Type>(ptf),
    regionCoupledPatch_(ptf.regionCoupledPatch_),
    nbrFieldName_(ptf.nbrFieldName_),
    matrixUpdateBuffer_()
{}


template<class Type>
Foam::regionCouplingFvPatchField<Type>::
regionCouplingFvPatchField
(
    const regionCouplingFvPatchField<Type>& ptf,
    const DimensionedField<Type, volMesh>& iF
)
:
    coupledFvPatchField<Type>(ptf, iF),
    regionCoupledPatch_(ptf.regionCoupledPatch_),
    nbrFieldName_(ptf.nbrFieldName_),
    matrixUpdateBuffer_()
{}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

template<class Type>
Foam::tmp<Foam::Field<Type> > Foam::regionCouplingFvPatchField<Type>::
snGrad() const
{
    return
        regionCoupledPatch_.patch().deltaCoeffs()
       *(*this - this->patchInternalField());
}


template<class Type>
Foam::tmp<Foam::Field<Type> > Foam::regionCouplingFvPatchField<Type>::
snGrad(const scalarField&) const
{
    return snGrad();
}


template<class Type>
void Foam::regionCouplingFvPatchField<Type>::evaluate
(
    const Pstream::commsTypes
)
{
    if (!this->updated())
    {
        this->updateCoeffs();
    }

    const scalarField lWeights(weights());

    Field<Type>::operator=
    (
            lWeights * this->patchInternalField()
         + (1.0 - lWeights) * this->patchNeighbourField()
    );

    fvPatchField<Type>::evaluate();
}


template<class Type>
const Foam::regionCouplingFvPatchField<Type>& Foam::regionCouplingFvPatchField<Type>::
neighbourPatchField() const
{
    // Lookup neighbour field
    typedef GeometricField<Type, fvPatchField, volMesh> GeoField;

    const GeoField& coupledField =
        regionCoupledPatch_.nbrMesh().
        objectRegistry::lookupObject<GeoField>(nbrFieldName_);

    return refCast<const regionCouplingFvPatchField<Type> >
    (
        coupledField.boundaryField()[regionCoupledPatch_.neighbPatchID()]
    );
}


template<class Type>
Foam::tmp<Foam::Field<Type> > Foam::regionCouplingFvPatchField<Type>::
patchNeighbourField() const
{
    return regionCoupledPatch_.regionCoupledPatch().interpolate
    (
        neighbourPatchField().patchInternalField()
    );
}


template<class Type>
void Foam::regionCouplingFvPatchField<Type>::initInterfaceMatrixUpdate
(
    scalarField&,
    const scalarField& psiInternal,
    const scalarField&,
    const direction,
    const Pstream::commsTypes
) const
{
    matrixUpdateBuffer_ = this->patch().patchInternalField(psiInternal);
}


template<class Type>
void Foam::regionCouplingFvPatchField<Type>::updateInterfaceMatrix
(
    Field<scalar>& result,
    const scalarField& psiInternal,
    const scalarField& coeffs,
    const direction cmpt,
    const Pstream::commsTypes
) const
{
    scalarField pnf = 
        regionCoupledPatch_.regionCoupledPatch().interpolate
        (
            this->neighbourPatchField().matrixUpdateBuffer()
        );

    // Multiply the field by coefficients and add into the result
    const labelUList& faceCells = regionCoupledPatch_.faceCells();

    forAll(faceCells, elemI)
    {
        result[faceCells[elemI]] -= coeffs[elemI]*pnf[elemI];
    }
}


template<class Type>
void Foam::regionCouplingFvPatchField<Type>::updateInterfaceMatrix
(
    Field<Type>& result,
    const Field<Type>& psiInternal,
    const scalarField& coeffs,
    const Pstream::commsTypes
) const
{
    notImplemented
    (
        "void Foam::regionCouplingFvPatchField<Type>::updateInterfaceMatrix"
        "("
            "Field<Type>&, "
            "const Field<Type>&, "
            "const scalarField&, "
            "const Pstream::commsTypes"
        ") const"
    );
}


template<class Type>
void Foam::regionCouplingFvPatchField<Type>::write(Ostream& os) const
{
    fvPatchField<Type>::write(os);
    os.writeKeyword("nbrField")
        << nbrFieldName_ << token::END_STATEMENT << nl;
    this->writeEntry("value", os);
}

// ************************************************************************* //
