/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright held by original author
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

Class
    Foam::regionCouplingFvPatchScalarField

Description
    Templated region coupled implicit boundary condition.
    The fvPatch is treated as uncoupled from the delta point of view.
    In the mesh the fvPatch is an interface and is incorporated
    into the matrix implicitly.

    The underlaying patch type should be regionCoupled, e.g.:
    \verbatim
    myPatch
    {
        type            regionCoupled;
        neighbourRegion nbrRegion;
        neighbourPatch  nbrPatch;
        nFaces          n;
        startFace       m;
    }
    \endverbatim

    Example of the boundary condition specification:
    \verbatim
    myPatch
    {
        type            regionCoupling;
        value           uniform (0 0 0);
        nbrField        U;
    }
    \endverbatim

Author
    Bahram Haddadi, TUWards OpenFOAM.  All rights reserved

SourceFiles
    regionCouplingFvPatchScalarField.C

\*---------------------------------------------------------------------------*/

#ifndef regionCouplingFvPatchField_H
#define regionCouplingFvPatchField_H

#include "regionCoupledBaseFvPatch.H"
#include "coupledFvPatchField.H"


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

/*---------------------------------------------------------------------------*\
             Class regionCouplingFvPatchScalarField Declaration
\*---------------------------------------------------------------------------*/

template<class Type>
class regionCouplingFvPatchField
:
    public coupledFvPatchField<Type>
{

private:

    // Private data

        //- Local reference to region couple patch
        const regionCoupledBaseFvPatch& regionCoupledPatch_;

        //- Name of nbr field to couple to
        word nbrFieldName_;

        //- Matrix update buffer
        mutable scalarField matrixUpdateBuffer_;


    // Private functions

        //- Local weight for this coupled field
        scalarField weights() const;

        //- Return nbr patch field
        const regionCouplingFvPatchField<Type>& neighbourPatchField() const;

        //- Return nbr field given internal cell data
        tmp<Field<Type> > patchNeighbourField() const;



protected:

        //- Return contents of a matrix update buffer
        const scalarField& matrixUpdateBuffer() const
        {
            return matrixUpdateBuffer_;
        }


public:

    //- Runtime type information
    TypeName("regionCoupling");


    // Constructors

        //- Construct from patch and internal field
        regionCouplingFvPatchField
        (
            const fvPatch&,
            const DimensionedField<Type, volMesh>&
        );

        //- Construct from patch, internal field and dictionary
        regionCouplingFvPatchField
        (
            const fvPatch&,
            const DimensionedField<Type, volMesh>&,
            const dictionary&
        );

        //- Construct by mapping given regionCouplingFvPatchScalarField
        // onto a new patch
        regionCouplingFvPatchField
        (
            const regionCouplingFvPatchField<Type>&,
            const fvPatch&,
            const DimensionedField<Type, volMesh>&,
            const fvPatchFieldMapper&
        );

        //- Construct as copy
        regionCouplingFvPatchField
        (
            const regionCouplingFvPatchField<Type>&
        );

        //- Construct and return a clone
        virtual tmp<fvPatchField<Type> > clone() const
        {
            return tmp<fvPatchField<Type> >
            (
                new regionCouplingFvPatchField<Type>(*this)
            );
        }

        //- Construct as copy setting internal field reference
        regionCouplingFvPatchField
        (
            const regionCouplingFvPatchField<Type>&,
            const DimensionedField<Type, volMesh>&
        );

        //- Construct and return a clone setting internal field reference
        virtual tmp<fvPatchField<Type> > clone
        (
            const DimensionedField<Type, volMesh>& iF
        ) const
        {
            return tmp<fvPatchField<Type> >
            (
                new regionCouplingFvPatchField<Type>(*this, iF)
            );
        }


    //- Destructor
    virtual ~regionCouplingFvPatchField()
    {}

        // Evaluation functions

            //- Return nbr field name
            const word& nbrFieldName() const
            {
                return nbrFieldName_;
            }

            //- Return patch-normal gradient
            virtual tmp<Field<Type> > snGrad() const;

            //- Return patch-normal gradient
            //  Note: the deltaCoeffs supplied are not used
            virtual tmp<Field<Type> > snGrad
            (
                const scalarField& deltaCoeffs
            ) const;


            //- Evaluate the patch field
            virtual void evaluate
            (
                const Pstream::commsTypes commsType
            );


        // Coupled interface functionality

            //- Initialise neighbour matrix update
            virtual void initInterfaceMatrixUpdate
            (
                scalarField& result,
                const scalarField& psiInternal,
                const scalarField& coeffs,
                const direction cmpt,
                const Pstream::commsTypes commsType
            ) const;

            //- Update result field based on interface functionality
            virtual void updateInterfaceMatrix
            (
                Field<scalar>& result,
                const scalarField& psiInternal,
                const scalarField& coeffs,
                const direction cmpt,
                const Pstream::commsTypes commsType
            ) const;

            //- Update result field based on interface functionality
            virtual void updateInterfaceMatrix
            (
                Field<Type>&,
                const Field<Type>&,
                const scalarField&,
                const Pstream::commsTypes commsType
            ) const;

            //- Return the interface type
            virtual const word& interfaceFieldType() const
            {
                return regionCoupledPatch_.regionCoupleType();
            }


        //- Write
        virtual void write(Ostream&) const;
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#ifdef NoRepository
#   include "regionCouplingFvPatchField.C"
#endif


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
