/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright held by original author
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

Description
    Coupled Finite Volume matrix

\*---------------------------------------------------------------------------*/

#include "coupledLduSolver.H"

// * * * * * * * * * * * * * Private Member Functions  * * * * * * * * * * * //

template<class Type>
void Foam::coupledFvMatrix<Type>::checkSize() const
{
    if (this->size() < 1)
    {
        FatalErrorIn("void Foam::coupledFvMatrix<Type>::checkSize() const")
            << "No matrices added to coupled matrix"
            << abort(FatalError);
    }
}


template<class Type>
Foam::word Foam::coupledFvMatrix<Type>::coupledPsiName() const
{
    word cpn;

    const UPtrList<lduMatrix>& matrices = *this;

    const fvMatrix<Type>& curMatrix =
        static_cast<const fvMatrix<Type>& >(matrices[0]);

    cpn = "coupled_" + curMatrix.psi().name();

    return cpn;
}

// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

template<class Type>
Foam::coupledSolverPerformance
Foam::coupledFvMatrix<Type>::solve(const dictionary& solverControls)
{
    if (debug)
    {
        InfoIn("coupledFvMatrix<Type>::solve(const dictionary&)")
            << "solving coupledFvMatrix<Type>" << endl;
    }

    coupledSolverPerformance solverPerfVec
    (
        "coupledFvMatrix<Type>::solve",
        this->coupledPsiName()
    );

    typedef FieldField<Field, scalar> scalarFieldField;

    UPtrList<lduMatrix>& matrices = *this;

    // Make a copy of the diagonal and complete the source
    scalarFieldField saveDiag(this->size());
    scalarFieldField psiCmpt(this->size());
    FieldField<Field, Type> source(this->size());
    scalarFieldField sourceCmpt(this->size());
    lduInterfaceFieldPtrsListList interfaces(this->size());

    // Prepare block solution
    forAll (matrices, rowI)
    {
        fvMatrix<Type>& curMatrix =
            static_cast<fvMatrix<Type>& >(matrices[rowI]);

        saveDiag.set(rowI, new scalarField(curMatrix.diag()));
        psiCmpt.set(rowI, new scalarField(curMatrix.psi().size()));
        source.set(rowI, new Field<Type>(curMatrix.source()));
        sourceCmpt.set(rowI, new scalarField(curMatrix.psi().size()));

        addBoundarySource(source[rowI], 0, curMatrix, rowI);

        interfaces[rowI] = curMatrix.psi().boundaryField().scalarInterfaces();
    }

    for (direction cmpt = 0; cmpt < Type::nComponents; cmpt++)
    {
        // Copy field and source

        PtrList<FieldField<Field, scalar> > bouCoeffsCmpt(this->size());
        PtrList<FieldField<Field, scalar> > intCoeffsCmpt(this->size());

        forAll (matrices, rowI)
        {

            fvMatrix<Type>& curMatrix =
                static_cast<fvMatrix<Type>& >(matrices[rowI]);

            psiCmpt[rowI] = curMatrix.psi().internalField().component(cmpt);
            addBoundaryDiag(curMatrix.diag(), cmpt, curMatrix, rowI);

            sourceCmpt[rowI] = source[rowI].component(cmpt);

            bouCoeffsCmpt.set
            (
                rowI,
                new FieldField<Field, scalar>
                (
                    curMatrix.boundaryCoeffs().component(cmpt)
                )
            );

            intCoeffsCmpt.set
            (
                rowI,
                new FieldField<Field, scalar>
                (
                    curMatrix.internalCoeffs().component(cmpt)
                )
            );
        }

        this->initMatrixInterfaces
        (
            bouCoeffsCmpt,
            interfaces,
            psiCmpt,
            sourceCmpt,
            cmpt
        );

        this->updateMatrixInterfaces
        (
            bouCoeffsCmpt,
            interfaces,
            psiCmpt,
            sourceCmpt,
            cmpt
        );

        coupledSolverPerformance solverPerf
        (
            "coupledFvMatrix<Type>::solve",
            this->coupledPsiName()
        );

        // Solver call
        solverPerf = coupledLduSolver::New
        (
            this->coupledPsiName() + pTraits<Type>::componentNames[cmpt],
            *this,
            bouCoeffsCmpt,
            intCoeffsCmpt,
            interfaces,
            solverControls
        )->solve(psiCmpt, sourceCmpt, cmpt);

        solverPerf.print();

        if
        (
            solverPerf.initialResidual() > solverPerfVec.initialResidual()
         && !solverPerf.singular()
        )
        {
            solverPerfVec = solverPerf;
        }

        // Update solution
        forAll (matrices, rowI)
        {
            fvMatrix<Type>& curMatrix =
                static_cast<fvMatrix<Type>& >(matrices[rowI]);
            const_cast<GeometricField<Type, fvPatchField, volMesh>&>
            (curMatrix.psi()).internalField().replace(cmpt, psiCmpt[rowI]);
            curMatrix.diag() = saveDiag[rowI];
        }
    }

    // Correct boundart conditions
    forAll (matrices, rowI)
    {
        fvMatrix<Type>& curMatrix =
            static_cast<fvMatrix<Type>& >(matrices[rowI]);
        const_cast<GeometricField<Type, fvPatchField, volMesh>&>
        (curMatrix.psi()).correctBoundaryConditions();
    }

    return solverPerfVec;
}


template<class Type>
Foam::coupledSolverPerformance Foam::coupledFvMatrix<Type>::solve()
{
    // Check matrices are present.  Using first matrix for controls
    checkSize();

    const fvMatrix<Type>& m =
        static_cast<const fvMatrix<Type>& >(this->operator[](0));

    return solve(m.psi().mesh().solver(coupledPsiName()));
}

template<class Type>
inline Foam::fvMatrix<Type>* Foam::coupledFvMatrix<Type>::setEqn(const label i, fvMatrix<Type>* ptr)
{
    // Keeping a copy for deallocating memory
    ptrs_.set(i, ptr);

    // Accessing PtrList set function
    this->set(i, ptr);

    return ptr;
}

template<class Type>
void Foam::coupledFvMatrix<Type>::relax()
{
    forAll(*this, i)
    {
        ptrs_[i].relax();
    }
}

template<class Type>
template<class Type2>
void Foam::coupledFvMatrix<Type>::addToInternalField
(
    const labelUList& addr,
    const Field<Type2>& pf,
    Field<Type2>& intf
) const
{
    if (addr.size() != pf.size())
    {
        FatalErrorIn
        (
            "fvMatrix<Type>::addToInternalField(const labelUList&, "
            "const Field&, Field&)"
        )   << "sizes of addressing and field are different"
            << abort(FatalError);
    }

    forAll(addr, faceI)
    {
        intf[addr[faceI]] += pf[faceI];
    }
}

template<class Type>
template<class Type2>
void Foam::coupledFvMatrix<Type>::addToInternalField
(
    const labelUList& addr,
    const tmp<Field<Type2> >& tpf,
    Field<Type2>& intf
) const
{
    addToInternalField(addr, tpf(), intf);
    tpf.clear();
}

template<class Type>
void Foam::coupledFvMatrix<Type>::addBoundaryDiag
(
    scalarField& diag,
    const direction solveCmpt,
    fvMatrix<Type>& curMatrix,
    const scalar rowI
) const
{
    FieldField<Field, Type>& internalCoeffs_ = curMatrix.internalCoeffs();

    forAll(internalCoeffs_, patchI)
    {
        addToInternalField
        (
            lduAddr(rowI).patchAddr(patchI),
            internalCoeffs_[patchI].component(solveCmpt),
            diag
        );
    }
}

template<class Type>
void Foam::coupledFvMatrix<Type>::addBoundarySource
(
    Field<Type>& source,
    const bool couples,
    fvMatrix<Type>& curMatrix,
    const scalar rowI
) const
{
    const GeometricField<Type, fvPatchField, volMesh>& psi_ = curMatrix.psi();
    FieldField<Field, Type>& boundaryCoeffs_ = curMatrix.boundaryCoeffs();

    forAll(psi_.boundaryField(), patchI)
    {
        const fvPatchField<Type>& ptf = psi_.boundaryField()[patchI];
        const Field<Type>& pbc = boundaryCoeffs_[patchI];

        if (!ptf.coupled())
        {
            addToInternalField(lduAddr(rowI).patchAddr(patchI), pbc, source);
        }
        else if (couples)
        {
            tmp<Field<Type> > tpnf = ptf.patchNeighbourField();
            const Field<Type>& pnf = tpnf();

            const labelUList& addr = lduAddr(rowI).patchAddr(patchI);

            forAll(addr, facei)
            {
                source[addr[facei]] += cmptMultiply(pbc[facei], pnf[facei]);
            }
        }
    }
}

// ************************************************************************* //
