/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright held by original author
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

Description
     Finite-Volume scalar matrix member functions and operators

\*---------------------------------------------------------------------------*/

#include "lduInterfaceFieldPtrsList.H"
#include "coupledFvScalarMatrix.H"
#include "zeroGradientFvPatchFields.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

template<>
coupledSolverPerformance coupledFvMatrix<scalar>::solve
(
    const dictionary& solverControls
)
{
    if (debug)
    {
        InfoIn("coupledFvMatrix<Type>::solve(const dictionary)")
            << "solving coupledFvMatrix<Type>" << endl;
    }

    typedef FieldField<Field, scalar> scalarFieldField;

    typedef List<lduInterfaceFieldPtrsList> lduInterfaceFieldPtrsListList;

    UPtrList<lduMatrix>& matrices = *this;

    // Make a copy of the diagonal and complete the source
    scalarFieldField saveDiag(this->size());
    scalarFieldField psi(this->size());
    FieldField<Field, scalar> source(this->size());
    lduInterfaceFieldPtrsListList interfaces(this->size());

    PtrList<FieldField<Field, scalar> > bouCoeffs(this->size());
    PtrList<FieldField<Field, scalar> > intCoeffs(this->size());

    // Prepare block solution
    forAll (matrices, rowI)
    {
        fvScalarMatrix& curMatrix =
            static_cast<fvScalarMatrix&>(matrices[rowI]);

        saveDiag.set(rowI, new scalarField(curMatrix.diag()));
        psi.set(rowI, new scalarField(curMatrix.psi()));
        source.set(rowI, new scalarField(curMatrix.source()));

        addBoundarySource(source[rowI], 0, curMatrix, rowI);

        interfaces[rowI] = curMatrix.psi().boundaryField().scalarInterfaces();

        addBoundaryDiag(curMatrix.diag(), 0, curMatrix, rowI);

        bouCoeffs.set
        (
            rowI,
            new FieldField<Field, scalar>
            (
                curMatrix.boundaryCoeffs()
            )
        );

        intCoeffs.set
        (
            rowI,
            new FieldField<Field, scalar>
            (
                curMatrix.internalCoeffs().component(0)
            )
        );
    }

    // Solver call
    coupledSolverPerformance solverPerf = coupledLduSolver::New
    (
        this->coupledPsiName(),
        *this,
        bouCoeffs,
        intCoeffs,
        interfaces,
        solverControls
    )->solve(psi, source, 0);

    solverPerf.print();

    // Update solution
    forAll (matrices, rowI)
    {
        fvScalarMatrix& curMatrix =
            static_cast<fvScalarMatrix&>(matrices[rowI]);

        // Nasty part :(
        const_cast<GeometricField<scalar, fvPatchField, volMesh>&>
        (curMatrix.psi()).internalField() = psi[rowI];
    }

    // Correct boundary conditions
    forAll (matrices, rowI)
    {
        fvScalarMatrix& curMatrix =
            static_cast<fvScalarMatrix&>(matrices[rowI]);

        // Nasty part :(
        const_cast<GeometricField<scalar, fvPatchField, volMesh>&>
        (curMatrix.psi()).correctBoundaryConditions();
    }

    return solverPerf;
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// ************************************************************************* //
