/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright held by original author
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

Class
    coupledLduMatrix

Description
    Collection of lduMatrices solved together as a block system

Author
    Hrvoje Jasak, Wikki Ltd.  All rights reserved

Rewriting for OpenFOAM 2.3.0
    Bahram Haddadi, TUWards OpenFOAM.  All rights reserved

\*----------------------------------------------------------------------------*/

#include "coupledLduMatrix.H"

// * * * * * * * * * * * * * * Static Data Members * * * * * * * * * * * * * //

namespace Foam
{
    defineTypeNameAndDebug(coupledLduMatrix, 1);
}


// * * * * * * * * * * * * * Private Member Functions  * * * * * * * * * * * //


// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

// Construct given size
Foam::coupledLduMatrix::coupledLduMatrix(const label size)
:
    UPtrList<lduMatrix>(size)
{}


// * * * * * * * * * * * * * * * * Destructor  * * * * * * * * * * * * * * * //

Foam::coupledLduMatrix::~coupledLduMatrix()
{}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

const Foam::lduMesh& Foam::coupledLduMatrix::mesh(const label rowI) const
{
    const UPtrList<lduMatrix>& matrices = *this;

    return matrices[rowI].mesh();
}

const Foam::lduAddressing& Foam::coupledLduMatrix::lduAddr(const label rowI) const
{
    const UPtrList<lduMatrix>& matrices = *this;

    return matrices[rowI].lduAddr();
}

const Foam::lduSchedule& Foam::coupledLduMatrix::patchSchedule(const label rowI) const
{
    const UPtrList<lduMatrix>& matrices = *this;

    return matrices[rowI].patchSchedule();
}

Foam::scalarField& Foam::coupledLduMatrix::lower(const label rowI)
{
    UPtrList<lduMatrix>& matrices = *this;

    return matrices[rowI].lower();
}

Foam::scalarField& Foam::coupledLduMatrix::diag(const label rowI)
{
    UPtrList<lduMatrix>& matrices = *this;

    return matrices[rowI].diag();
}

Foam::scalarField& Foam::coupledLduMatrix::upper(const label rowI)
{
    UPtrList<lduMatrix>& matrices = *this;

    return matrices[rowI].upper();
}

const Foam::scalarField& Foam::coupledLduMatrix::lower(const label rowI) const
{
    const UPtrList<lduMatrix>& matrices = *this;

    return matrices[rowI].lower();
}

const Foam::scalarField& Foam::coupledLduMatrix::diag(const label rowI) const
{
    const UPtrList<lduMatrix>& matrices = *this;

    return matrices[rowI].diag();
}

const Foam::scalarField& Foam::coupledLduMatrix::upper(const label rowI) const
{
    const UPtrList<lduMatrix>& matrices = *this;

    return matrices[rowI].upper();
}

bool Foam::coupledLduMatrix::hasDiag(const label rowI) const
{
    const UPtrList<lduMatrix>& matrices = *this;

    return matrices[rowI].hasDiag();
}

bool Foam::coupledLduMatrix::hasUpper(const label rowI) const
{
    const UPtrList<lduMatrix>& matrices = *this;

    return matrices[rowI].hasUpper();
}

bool Foam::coupledLduMatrix::hasLower(const label rowI) const
{
    const UPtrList<lduMatrix>& matrices = *this;

    return matrices[rowI].hasLower();
}

bool Foam::coupledLduMatrix::diagonal() const
{
    const UPtrList<lduMatrix>& matrices = *this;

    bool diag = true;

    forAll (matrices, matrixI)
    {
        diag = diag && matrices[matrixI].diagonal();
    }

    return diag;
}

bool Foam::coupledLduMatrix::symmetric() const
{
    const UPtrList<lduMatrix>& matrices = *this;

    bool sym = true;

    forAll (matrices, matrixI)
    {
        sym =
            (sym && matrices[matrixI].diagonal())
         || (sym && matrices[matrixI].symmetric());
    }

    return sym;
}

bool Foam::coupledLduMatrix::asymmetric() const
{
    const UPtrList<lduMatrix>& matrices = *this;

    bool asym = false;

    forAll (matrices, matrixI)
    {
        asym = (asym || matrices[matrixI].asymmetric());
    }

    return asym;
}

void Foam::coupledLduMatrix::AmulCore
(
    scalarField& Ax,
    const scalarField& x,
    const label rowI
) const
{
    scalar* __restrict__ AxPtr = Ax.begin();

    const scalar* const __restrict__ xPtr = x.begin();

    // Protection for multiplication of incomplete matrices
    if (hasDiag(rowI))
    {
        const scalar* const __restrict__ diagPtr = diag(rowI).begin();

        register const label nCells = diag(rowI).size();
        for (register label cell=0; cell<nCells; cell++)
        {
            // AmulCore must be additive to account for initialisation step
            // in ldu interfaces.
            AxPtr[cell] += diagPtr[cell]*xPtr[cell];
        }
    }

    // Protection for multiplication of incomplete matrices
    if (hasUpper(rowI) || hasLower(rowI))
    {
        const label* const __restrict__ uPtr = lduAddr(rowI).upperAddr().begin();
        const label* const __restrict__ lPtr = lduAddr(rowI).lowerAddr().begin();

        const scalar* const __restrict__ upperPtr = upper(rowI).begin();
        const scalar* const __restrict__ lowerPtr = lower(rowI).begin();

        register const label nFaces = upper(rowI).size();

        for (register label face=0; face<nFaces; face++)
        {
            AxPtr[uPtr[face]] += lowerPtr[face]*xPtr[lPtr[face]];
            AxPtr[lPtr[face]] += upperPtr[face]*xPtr[uPtr[face]];
        }
    }
}

void Foam::coupledLduMatrix::Amul
(
    FieldField<Field, scalar>& result,
    const FieldField<Field, scalar>& x,
    const PtrList<FieldField<Field, scalar> >& bouCoeffs,
    const lduInterfaceFieldPtrsListList& interfaces,
    const direction cmpt
) const
{
    const UPtrList<lduMatrix>& matrices = *this;

    // Reset product to zero
    result = 0;

    // Initialise the update of coupled interfaces
    initMatrixInterfaces
    (
        bouCoeffs,
        interfaces,
        x,
        result,
        cmpt
    );

    forAll (matrices, rowI)
    {
        AmulCore(result[rowI], x[rowI], rowI);
    }

    // Update couple interfaces
    updateMatrixInterfaces
    (
        bouCoeffs,
        interfaces,
        x,
        result,
        cmpt
    );
}

void Foam::coupledLduMatrix::TmulCore
(
    scalarField& Tx,
    const scalarField& x,
    const label rowI
) const
{
    scalar* __restrict__ TxPtr = Tx.begin();

    const scalar* const __restrict__ xPtr = x.begin();

    // Protection for multiplication of incomplete matrices
    if (hasDiag(rowI))
    {
        const scalar* const __restrict__ diagPtr = diag(rowI).begin();

        register const label nCells = diag(rowI).size();
        for (register label cell=0; cell<nCells; cell++)
        {
            // TmulCore must be additive to account for initialisation step
            // in ldu interfaces.
            TxPtr[cell] += diagPtr[cell]*xPtr[cell];
        }
    }

    // Protection for multiplication of incomplete matrices
    if (hasUpper(rowI) || hasLower(rowI))
    {
        const label* const __restrict__ uPtr = lduAddr(rowI).upperAddr().begin();
        const label* const __restrict__ lPtr = lduAddr(rowI).lowerAddr().begin();

        const scalar* const __restrict__ lowerPtr = lower(rowI).begin();
        const scalar* const __restrict__ upperPtr = upper(rowI).begin();

        register const label nFaces = upper(rowI).size();
        for (register label face=0; face<nFaces; face++)
        {
            TxPtr[uPtr[face]] += upperPtr[face]*xPtr[lPtr[face]];
            TxPtr[lPtr[face]] += lowerPtr[face]*xPtr[uPtr[face]];
        }
    }
}

void Foam::coupledLduMatrix::Tmul
(
    FieldField<Field, scalar>& result,
    const FieldField<Field, scalar>& x,
    const PtrList<FieldField<Field, scalar> >& intCoeffs,
    const lduInterfaceFieldPtrsListList& interfaces,
    const direction cmpt
) const
{
    const UPtrList<lduMatrix>& matrices = *this;

    // Reset product to zero
    result = 0;

    // Initialise the update of coupled interfaces
    initMatrixInterfaces
    (
        intCoeffs,
        interfaces,
        x,
        result,
        cmpt
    );

    forAll (matrices, rowI)
    {
        TmulCore(result[rowI], x[rowI], rowI);
    }

    // Update couple interfaces
    updateMatrixInterfaces
    (
        intCoeffs,
        interfaces,
        x,
        result,
        cmpt
    );
}


void Foam::coupledLduMatrix::initMatrixInterfaces
(
    const PtrList<FieldField<Field, scalar> >& coupleCoeffs,
    const lduInterfaceFieldPtrsListList& interfaces,
    const FieldField<Field, scalar>& x,
    FieldField<Field, scalar>& result,
    const direction cmpt
) const
{
    const UPtrList<lduMatrix>& matrices = *this;

    forAll (matrices, rowI)
    {
        matrices[rowI].initMatrixInterfaces
        (
            coupleCoeffs[rowI],
            interfaces[rowI],
            x[rowI],
            result[rowI],
            cmpt
        );
    }
}


void Foam::coupledLduMatrix::updateMatrixInterfaces
(
    const PtrList<FieldField<Field, scalar> >& coupleCoeffs,
    const lduInterfaceFieldPtrsListList& interfaces,
    const FieldField<Field, scalar>& x,
    FieldField<Field, scalar>& result,
    const direction cmpt
) const
{
    const UPtrList<lduMatrix>& matrices = *this;

    forAll (matrices, rowI)
    {
        matrices[rowI].updateMatrixInterfaces
        (
            coupleCoeffs[rowI],
            interfaces[rowI],
            x[rowI],
            result[rowI],
            cmpt
        );
    }
}


// * * * * * * * * * * * * * * * Member Operators  * * * * * * * * * * * * * //


// * * * * * * * * * * * * * * * Friend Functions  * * * * * * * * * * * * * //


// * * * * * * * * * * * * * * * Friend Operators  * * * * * * * * * * * * * //


// ************************************************************************* //
